@@ -1,4 +1,5 @@
-/* Copyright (c) 2003,2004,2005 David Lichteblau
+/* -*- show-trailing-whitespace: t; indent-tabs: t -*-
+ * Copyright (c) 2003,2004,2005,2006 David Lichteblau
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -62,7 +63,7 @@ entry_free(tentry *entry)
 	GPtrArray *attributes = entry_attributes(entry);
 	int n = attributes->len;
 	int i;
-	
+
 	for (i = 0; i < n; i++)
 		attribute_free(g_ptr_array_index(attributes, i));
 	named_array_free((named_array *) entry);
@@ -93,7 +94,7 @@ attribute_free(tattribute *attribute)
 	GPtrArray *values = attribute_values(attribute);
 	int n = values->len;
 	int i;
-	
+
 	for (i = 0; i < n; i++)
 		g_array_free(g_ptr_array_index(values, i), 1);
 	named_array_free((named_array *) attribute);
@@ -161,14 +162,69 @@ attribute_remove_value(tattribute *a, char *data, int n)
 	return 0;
 }
 
+/*
+ * Aus irgendwelchen Gruenden habe ich mal beschlossen, GArrays mit chars drin
+ * statt GStrings zu nehmen fuer die Attributwerte.  Wie unpraktisch.
+ */
+char *
+array2string(GArray *av)
+{
+	int n = av->len;
+	char *str = xalloc(n + 1);
+	memcpy(str, av->data, n);
+	str[n] = 0;
+	return str;
+}
+
+/*
+ * allocate a new berval and copy LEN bytes of DATA into it
+ */
+static struct berval *
+dup2berval(char *data, int len)
+{
+	struct berval *bv = xalloc(sizeof(struct berval));
+	bv->bv_val = xalloc(len);
+	memcpy(bv->bv_val, data, len);
+	bv->bv_len = len;
+	return bv;
+}
+
+void
+xfree_berval(struct berval *bv)
+{
+	free(bv->bv_val);
+	free(bv);
+}
+
 struct berval *
 string2berval(GArray *s)
 {
-	struct berval *bv = xalloc(sizeof(struct berval));
-	bv->bv_val = xalloc(s->len);
-	memcpy(bv->bv_val, s->data, s->len);
-	bv->bv_len = s->len;
-	return bv;
+	return dup2berval(s->data, s->len);
+}
+
+struct berval *
+gstring2berval(GString *s)
+{
+	return dup2berval(s->str, s->len);
+}
+
+LDAPMod *
+attribute2mods(tattribute *attribute)
+{
+	GPtrArray *values = attribute_values(attribute);
+	LDAPMod *m = xalloc(sizeof(LDAPMod));
+	int j;
+
+	m->mod_op = LDAP_MOD_BVALUES;
+	m->mod_type = xdup(attribute_ad(attribute));
+	m->mod_bvalues = xalloc(
+		(1 + values->len) * sizeof(struct berval *));
+
+	for (j = 0; j < values->len; j++)
+		m->mod_bvalues[j]
+			= string2berval(g_ptr_array_index(values, j));
+	m->mod_bvalues[j] = 0;
+	return m;
 }
 
 LDAPMod **
@@ -176,24 +232,10 @@ entry2mods(tentry *entry)
 {
 	GPtrArray *attributes = entry_attributes(entry);
 	LDAPMod **result = xalloc((attributes->len + 1) * sizeof(LDAPMod *));
-	int i, j;
+	int i;
 
-	for (i = 0; i < attributes->len; i++) {
-		tattribute *attribute = g_ptr_array_index(attributes, i);
-		GPtrArray *values = attribute_values(attribute);
-		LDAPMod *m = xalloc(sizeof(LDAPMod));
-
-		m->mod_op = LDAP_MOD_BVALUES;
-		m->mod_type = xdup(attribute_ad(attribute));
-		m->mod_bvalues = xalloc(
-			(1 + values->len) * sizeof(struct berval *));
-
-		for (j = 0; j < values->len; j++)
-			m->mod_bvalues[j]
-				= string2berval(g_ptr_array_index(values, j));
-		m->mod_bvalues[j] = 0;
-		result[i] = m;
-	}
+	for (i = 0; i < attributes->len; i++)
+		result[i] = attribute2mods(g_ptr_array_index(attributes, i));
 	result[i] = 0;
 	return result;
 }
