@@ -1,4 +1,5 @@
-/* Copyright (c) 2003,2004,2005 David Lichteblau
+/* -*- show-trailing-whitespace: t; indent-tabs: t -*-
+ * Copyright (c) 2003,2004,2005,2006 David Lichteblau
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -15,6 +16,7 @@
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 #include "common.h"
+#include "config.h"
 
 typedef void (*note_function)(void *, void *, void *);
 
@@ -45,42 +47,24 @@ compare_ptr_arrays(GPtrArray *a, GPtrArray *b,
 }
 
 static void
-note_values(GArray *a, GArray *b, GPtrArray *mods)
+note_values(GArray *a, GArray *b, int *changed)
 {
-	GArray *v;
-	LDAPMod *m;
-
-	if (a && b) return;
-
-	m = xalloc(sizeof(LDAPMod));
-	if (a) {
-		v = a;
-		m->mod_op = LDAP_MOD_DELETE;
-	} else {
-		v = b;
-		m->mod_op = LDAP_MOD_ADD;
-	}
-	m->mod_op |= LDAP_MOD_BVALUES;
-	m->mod_type = 0;
-	m->mod_bvalues = xalloc(2 * sizeof(struct berval *));
-	m->mod_bvalues[0] = string2berval(v);
-	m->mod_bvalues[1] = 0;
-	g_ptr_array_add(mods, m);
+	if (!(a && b)) *changed = 1;
 }
 
 static void
 compare_attributes(tattribute *clean, tattribute *new, GPtrArray *mods)
 {
-	int i;
+	int changed = 0;
 	compare_ptr_arrays(attribute_values(clean),
 			   attribute_values(new),
 			   carray_ptr_cmp,
 			   (note_function) note_values,
-			   mods);
-	for (i = mods->len - 1; i >= 0; i--) {
-		LDAPMod *m = g_ptr_array_index(mods, i);
-		if (m->mod_type) break;
-		m->mod_type = xdup(attribute_ad(clean));
+			   &changed);
+	if (changed) {
+		LDAPMod *m = attribute2mods(new);
+		m->mod_op |= LDAP_MOD_REPLACE;
+		g_ptr_array_add(mods, m);
 	}
 }
 
@@ -91,7 +75,7 @@ note_attributes(tattribute *a1, tattribute *a2, GPtrArray *mods)
 	GPtrArray *values;
 	LDAPMod *m;
 	int i;
-	
+
 	if (a1 && a2) {
 		compare_attributes(a1, a2, mods);
 		return;
@@ -117,7 +101,7 @@ note_attributes(tattribute *a1, tattribute *a2, GPtrArray *mods)
 	g_ptr_array_add(mods, m);
 }
 
-LDAPMod **
+static LDAPMod **
 compare_entries(tentry *eclean, tentry *enew)
 {
 	GPtrArray *mods = g_ptr_array_new();
@@ -216,6 +200,10 @@ frob_ava(tentry *entry, int mode, char *ad, char *data, int n)
 	return 0;
 }
 
+#if defined(LIBLDAP21)
+#warning compiling for libldap <= 2.1, running with >= 2.2 will result in segfault
+#define safe_str2dn ldap_str2dn
+#elif defined(LIBLDAP22)
 /*
  * the following is exactly equivalent to ldap_str2dn in libldap >= 2.2,
  * but will fail linking on 2.1.  This way we avoid calling the old 2.1
@@ -229,6 +217,9 @@ safe_str2dn(char *str, LDAPDN *out, int flags)
         bv.bv_len = strlen(str);
         ldap_bv2dn_x(&bv, out, flags);
 }
+#else
+#error oops
+#endif
 
 /*
  * Call frob_ava for every ava in DN's (first) RDN.
@@ -239,14 +230,22 @@ safe_str2dn(char *str, LDAPDN *out, int flags)
 int
 frob_rdn(tentry *entry, char *dn, int mode)
 {
+#ifdef LIBLDAP21
+	LDAPDN *olddn;
+#else
 	LDAPDN olddn;
+#endif
 	LDAPRDN rdn;
 	int i;
 	int rc = 0;
 
 	safe_str2dn(dn, &olddn, LDAP_DN_FORMAT_LDAPV3);
 
+#ifdef LIBLDAP21
+	rdn = (**olddn)[0];
+#else
 	rdn = olddn[0];
+#endif
 	for (i = 0; rdn[i]; i++) {
 		LDAPAVA *ava = rdn[i];
 		char *ad = ava->la_attr.bv_val; /* XXX */
@@ -300,7 +299,7 @@ validate_rename(tentry *clean, tentry *data, int *deleteoldrdn)
 	return 0;
 }
 
-void
+static void
 rename_entry(tentry *entry, char *newdn, int deleteoldrdn)
 {
 	if (deleteoldrdn)
@@ -310,16 +309,212 @@ rename_entry(tentry *entry, char *newdn, int deleteoldrdn)
 	entry_dn(entry) = xdup(newdn);
 }
 
-void
-update_clean_copy(GArray *offsets, char *key, FILE *s, tentry *cleanentry)
+static void
+update_clean_copy(
+	GArray *offsets, char *key, FILE *s, tentry *cleanentry, tparser *p)
 {
 	long pos = fseek(s, 0, SEEK_END);
 	if (pos == -1) syserr();
 	g_array_index(offsets, long, atoi(key)) = ftell(s);
-	print_entry_object(s, cleanentry, key);
+	p->print(s, cleanentry, key, 0);
 }
 
+/*
+ * read a changerecord of type `key' from `data', handle it, and return
+ *    0 on success
+ *   -1 on syntax error
+ *   -2 on handler error
+ */
 int
+process_immediate(tparser *p, thandler *handler, void *userdata, FILE *data,
+		  long datapos, char *key)
+{
+	if (!strcmp(key, "add")) {
+		tentry *entry;
+		LDAPMod **mods;
+		if (p->entry(data, datapos, 0, &entry, 0) == -1)
+			return -1;
+		mods = entry2mods(entry);
+		if (handler->add(-1, entry_dn(entry), mods, userdata) == -1) {
+			ldap_mods_free(mods, 1);
+			entry_free(entry);
+			return -2;
+		}
+		ldap_mods_free(mods, 1);
+		entry_free(entry);
+		entry = 0;
+	} else if (!strcmp(key, "replace")) {
+		tentry *entry;
+		LDAPMod **mods;
+		int i;
+		if (p->entry(data, datapos, 0, &entry, 0) == -1)
+			return -1;
+		mods = entry2mods(entry);
+		for (i = 0; mods[i]; i++) {
+			LDAPMod *mod = mods[i];
+			mod->mod_op &= LDAP_MOD_BVALUES;
+			mod->mod_op |= LDAP_MOD_REPLACE;
+		}
+		if (handler->change(-1,
+				    entry_dn(entry),
+				    entry_dn(entry),
+				    mods,
+				    userdata) == -1) {
+			ldap_mods_free(mods, 1);
+			entry_free(entry);
+			return -2;
+		}
+		ldap_mods_free(mods, 1);
+		entry_free(entry);
+		entry = 0;
+	} else if (!strcmp(key, "rename")) {
+		char *dn1;
+		char *dn2;
+		int deleteoldrdn;
+		int rc;
+		if (p->rename(data, datapos, &dn1, &dn2, &deleteoldrdn) ==-1)
+			return -1;
+		rc = handler->rename0(-1, dn1, dn2, deleteoldrdn, userdata);
+		free(dn1);
+		free(dn2);
+		if (rc)
+			return -2;
+	} else if (!strcmp(key, "delete")) {
+		char *dn;
+		int rc;
+		if (p->delete(data, datapos, &dn) == -1)
+			return -1;
+		rc = handler->delete(-1, dn, userdata);
+		free(dn);
+		if (rc)
+			return -2;
+	} else if (!strcmp(key, "modify")) {
+		char *dn;
+		LDAPMod **mods;
+		if (p->modify(data, datapos, &dn, &mods) ==-1)
+			return -1;
+		if (handler->change(-1, dn, dn, mods, userdata) == -1) {
+			free(dn);
+			ldap_mods_free(mods, 1);
+			return -2;
+		}
+	} else {
+		fprintf(stderr, "Error: Invalid key: `%s'.\n", key);
+		return -1;
+	}
+	return 0;
+}
+
+/*
+ * read the next entry from `data', its clean copy from `clean', process
+ * them as described for compare_streams, and return
+ *    0 on success
+ *   -1 on syntax error
+ *   -2 on handler error
+ */
+static int
+process_next_entry(
+	tparser *p, thandler *handler, void *userdata, GArray *offsets,
+	FILE *clean, FILE *data, char *key, long datapos)
+{
+	tentry *entry = 0;
+	tentry *cleanentry = 0;
+	int rc = -1;
+	LDAPMod **mods;
+	long pos;
+	char *ptr;
+	int n;
+	int rename, deleteoldrdn;
+
+	/* find clean copy */
+	n = strtol(key, &ptr, 10);
+	if (*ptr)
+		return process_immediate(
+			p, handler, userdata, data, datapos, key);
+	if (n < 0 || n >= offsets->len) {
+		fprintf(stderr, "Error: Invalid key: `%s'.\n", key);
+		goto cleanup;
+	}
+	pos = g_array_index(offsets, long, n);
+	if (pos < 0) {
+		fprintf(stderr, "Error: Duplicate entry %d.\n", n);
+		goto cleanup;
+	}
+
+	/* find precise position */
+	if (p->entry(clean, pos, 0, 0, &pos) == -1) abort();
+	/* fast comparison */
+	if (n + 1 < offsets->len) {
+		long next = g_array_index(offsets, long, n + 1);
+		if (next >= 0
+		    && !fastcmp(clean, data, pos, datapos, next-pos+1))
+		{
+			datapos += next - pos;
+			long_array_invert(offsets, n);
+			if (fseek(data, datapos, SEEK_SET) == -1)
+				syserr();
+			return 0;
+		}
+	}
+
+	/* if we get here, a quick scan found a difference in the
+	 * files, so we need to read the entries and compare them */
+	if (p->entry(data, datapos, 0, &entry, 0) == -1)
+		goto cleanup;
+	if (p->entry(clean, pos, 0, &cleanentry, 0) == -1) abort();
+
+	/* compare and update */
+	if ( (rename = strcmp(entry_dn(cleanentry), entry_dn(entry)))){
+		if (validate_rename(cleanentry, entry, &deleteoldrdn)){
+			rc = -1;
+			goto cleanup;
+		}
+		if (handler->rename(n, entry_dn(cleanentry), entry, userdata)
+		    == -1)
+		{
+			rc = -2;
+			goto cleanup;
+		}
+		rename_entry(cleanentry, entry_dn(entry), deleteoldrdn);
+	}
+	if ( (mods = compare_entries(cleanentry, entry))) {
+		if (handler->change(n,
+				    entry_dn(cleanentry),
+				    entry_dn(entry),
+				    mods,
+				    userdata)
+		    == -1)
+		{
+			if (mods) ldap_mods_free(mods, 1);
+			if (rename)
+				update_clean_copy(
+					offsets, key, clean, cleanentry, p);
+			rc = -2;
+			goto cleanup;
+		}
+		ldap_mods_free(mods, 1);
+	}
+
+	/* mark as seen */
+	long_array_invert(offsets, n);
+
+	entry_free(entry);
+	entry = 0;
+	entry_free(cleanentry);
+	cleanentry = 0;
+	return 0;
+
+cleanup:
+	if (entry) {
+		if (*entry_dn(entry))
+			fprintf(stderr, "Error at: %s\n", entry_dn(entry));
+		entry_free(entry);
+	}
+	if (cleanentry) entry_free(cleanentry);
+	return rc;
+}
+
+static int
 nonleaf_action(tentry *entry, GArray *offsets, int n)
 {
 	int i;
@@ -352,6 +547,74 @@ more_deletions:
 		     "  ? -- this help");
 		goto more_deletions;
 	}
+
+	/* notreached */
+	return 0;
+}
+
+/*
+ * process deletions as described for compare_streams.
+ * return 0 on success, -2 else.
+ */
+static int
+process_deletions(tparser *p,
+		  thandler *handler,
+		  void *userdata,
+		  GArray *offsets,
+		  FILE *clean)
+{
+	tentry *cleanentry = 0;
+	long pos;
+	int n;
+	int ignore_nonleaf = 0;
+	int n_leaf;
+	int n_nonleaf;
+
+	do {
+		if (ignore_nonleaf)
+			printf("Retrying %d failed deletion%s...\n",
+			       n_nonleaf,
+			       n_nonleaf == 1 ? "" : "s");
+		n_leaf = 0;
+		n_nonleaf = 0;
+		for (n = 0; n < offsets->len; n++) {
+			if ( (pos = g_array_index(offsets, long, n)) < 0)
+				continue;
+			if (p->entry(clean, pos, 0, &cleanentry, 0) == -1)
+				abort();
+			switch (handler->delete(
+					n, entry_dn(cleanentry), userdata))
+			{
+			case -1:
+				entry_free(cleanentry);
+				return -2;
+			case -2:
+				if (ignore_nonleaf) {
+					printf("Skipping non-leaf entry: %s\n",
+					       entry_dn(cleanentry));
+					n_nonleaf++;
+					break;
+				}
+				switch (nonleaf_action(cleanentry,offsets,n)) {
+				case 0:
+					entry_free(cleanentry);
+					return -2;
+				case 2:
+					ignore_nonleaf = 1;
+					/* fall through */
+				case 1:
+					n_nonleaf++;
+				}
+				break;
+			default:
+				n_leaf++;
+				long_array_invert(offsets, n);
+			}
+			entry_free(cleanentry);
+		}
+	} while (ignore_nonleaf && n_nonleaf > 0 && n_leaf > 0);
+
+	return n_nonleaf ? -2 : 0;
 }
 
 /*
@@ -365,203 +628,79 @@ more_deletions:
  * and calling read_entry().
  *
  * File DATA, a modified copy of CLEAN may contain entries in any order,
- * which must be numbered or labeled "add".  If a key is a number, the
- * corresponding entry in CLEAN must exist, it is read and compared to
- * the modified copy.
+ * which must be numbered or labeled "add", "rename", or "modify".  If a
+ * key is a number, the corresponding entry in CLEAN must exist, it is
+ * read and compared to the modified copy.
  *
- * For each change, call handler with arguments described below.
- * Handlers must return 0 on success, or -1 on failure.  (As a special
- * case, return value -2 on a deletion indicates an attempt to delete a
- * non-leaf entry, which is non-fatal.)
+ * For each change, call the appropriate handler method with arguments
+ * described below.  Handler methods must return 0 on success, or -1 on
+ * failure.  (As a special case, return value -2 on a deletion indicates
+ * an attempt to delete a non-leaf entry, which is non-fatal.)
  *
  * For each new entry (labeled with "add"), call
- *   handler(NULL, new_entry, mods, USERDATA)
+ *   handler->add(dn, mods, USERDATA)
  * where MODS is a LDAPMod structure for the new entry.
  *
  * For each entry present in CLEAN but not DATA, call
- *   handler(removed_entry, 0, 0, USERDATA)
+ *   handler->delete(dn, USERDATA)
  * (This step can be repeated in the case of non-leaf entries.)
  *
  * For each entry present in both files, handler can be called two times.
  * If the distinguished names of the old and new entry disagree, call
- *   handler(old_entry, new_entry, 0, USERDATA)
+ *   handler->change(old_entry, new_entry, 0, USERDATA)
  * If there are additional changes to the attributes of the entry, call
- *   handler(renamed_entry, new_entry, mods, USERDATA)
+ *   handler->change(renamed_entry, new_entry, mods, USERDATA)
  * where RENAMED_ENTRY is a copy of the original entry, which accounts
  * for attribute modifications due to a possible RDN change (new RDN
  * component values have to be added, and old RDN values be removed),
  * and MODS describes the changes between RENAMED_ENTRY and NEW_ENTRY.
  *
+ * Entries labeled "delete" are changerecords for which the handler is
+ * called as described above.
+ *
+ * Entries labeled "rename" are changerecords with their own method,
+ * called as:
+ *   handler->rename(olddn, newdn, deleteoldrdn, USERDATA)
+ *
  * Return 0 on success, -1 on parse error, -2 on handler failure.
  *
  * If an error occured, *error_position is the offset in DATA after
  * which the erroneous entry can be found.
  */
 int
-compare_streams(int (*handler)(tentry *, tentry *, LDAPMod **, void *),
+compare_streams(tparser *p,
+		thandler *handler,
 		void *userdata,
-		GArray *offsets, FILE *clean, FILE *data,
+		GArray *offsets,
+		FILE *clean,
+		FILE *data,
 		long *error_position,
 		long *syntax_error_position)
 {
-	tentry *entry = 0;
-	tentry *cleanentry = 0;
 	char *key = 0;
 	int n;
-	long pos;
-	char *ptr;
-	LDAPMod **mods;
-	int rc = -1;
-	int n_leaf;
-	int n_nonleaf;
-	int ignore_nonleaf = 0;
+	int rc;
 
 	for (;;) {
 		long datapos;
-		int rename, deleteoldrdn;
 
-		/* look at updated entry */
+		/* read updated entry */
 		if (key) { free(key); key = 0; }
-		if (read_entry(data, -1, &key, 0, &datapos) == -1) goto cleanup;
+		if (p->peek(data, -1, &key, &datapos) == -1) goto cleanup;
 		*error_position = datapos;
 		if (!key) break;
 
-		/* new entry? */
-		if (!strcmp(key, "add")) {
-			if (read_entry(data, datapos, 0, &entry, 0) == -1)
-				goto cleanup;
-			mods = entry2mods(entry);
-			if (handler(0, entry, mods, userdata) == -1) {
-				ldap_mods_free(mods, 1);
-				rc = -2;
-				goto cleanup;
-			}
-			ldap_mods_free(mods, 1);
-			entry_free(entry);
-			entry = 0;
-			continue;
-		}
-
-		/* find clean copy */
-		n = strtol(key, &ptr, 10);
-		if (*ptr || n < 0 || n >= offsets->len) {
-			fprintf(stderr, "Error: Invalid key: `%s'.\n", key);
+		/* and do something with it */
+		if ( (rc = process_next_entry(
+			      p, handler, userdata, offsets, clean, data,
+			      key, datapos)))
 			goto cleanup;
-		}
-		pos = g_array_index(offsets, long, n);
-		if (pos < 0) {
-			fprintf(stderr, "Error: Duplicate entry %d.\n", n);
-			goto cleanup;
-		}
-
-		/* find precise position */
-		if (read_entry(clean, pos, 0, 0, &pos) == -1) abort();
-		/* fast comparison */
-		if (n + 1 < offsets->len) {
-			long next = g_array_index(offsets, long, n + 1);
-			if (next >= 0
-			    && !fastcmp(clean, data, pos, datapos, next-pos+1))
-			{
-				datapos += next - pos;
-				long_array_invert(offsets, n);
-				if (fseek(data, datapos, SEEK_SET) == -1)
-					syserr();
-				continue;
-			}
-		}
-
-		/* if we get here, a quick scan found a difference in the
-		 * files, so we need to read the entries and compare them */
-		if (read_entry(data, datapos, 0, &entry, 0) == -1)
-			goto cleanup;
-		if (read_entry(clean, pos, 0, &cleanentry, 0) == -1) abort();
-
-		/* compare and update */
-		if ( (rename = strcmp(entry_dn(cleanentry), entry_dn(entry)))){
-			if (validate_rename(cleanentry, entry, &deleteoldrdn)){
-				rc = -1;
-				goto cleanup;
-			}
-			if (handler(cleanentry, entry, 0, userdata) == -1) {
-				rc = -2;
-				goto cleanup;
-			}
-			rename_entry(
-				cleanentry, entry_dn(entry), deleteoldrdn);
-		}
-		if ( (mods = compare_entries(cleanentry, entry))) {
-			if (handler(cleanentry, entry, mods, userdata) == -1) {
-				if (mods) ldap_mods_free(mods, 1);
-				if (rename)
-					update_clean_copy(offsets, key, clean, cleanentry);
-				rc = -2;
-				goto cleanup;
-			}
-			ldap_mods_free(mods, 1);
-		}
-
-		/* mark as seen */
-		long_array_invert(offsets, n);
-
-		entry_free(entry);
-		entry = 0;
-		entry_free(cleanentry);
-		cleanentry = 0;
 	}
 	if ( (*error_position = ftell(data)) == -1) syserr();
 
-	/* find deleted entries */
-	do {
-		if (ignore_nonleaf)
-			printf("Retrying %d failed deletion%s...\n",
-			       n_nonleaf,
-			       n_nonleaf == 1 ? "" : "s");
-		n_leaf = 0;
-		n_nonleaf = 0;
-		for (n = 0; n < offsets->len; n++) {
-			if ( (pos = g_array_index(offsets, long, n)) < 0)
-				continue;
-			if (read_entry(clean, pos, 0, &cleanentry, 0) == -1)
-				abort();
-			switch (handler(cleanentry, 0, 0, userdata)) {
-			case -1:
-				rc = -2;
-				goto cleanup;
-			case -2:
-				if (ignore_nonleaf) {
-					printf("Skipping non-leaf entry: %s\n",
-					       entry_dn(cleanentry));
-					n_nonleaf++;
-					break;
-				}
-				switch (nonleaf_action(cleanentry,offsets,n)) {
-				case 0:
-					rc = -2;
-					goto cleanup;
-				case 2:
-					ignore_nonleaf = 1;
-					/* fall through */
-				case 1:
-					n_nonleaf++;
-				}
-				break;
-			default:
-				n_leaf++;
-				entry_free(cleanentry);
-				cleanentry = 0;
-				long_array_invert(offsets, n);
-			}
-		}
-	} while (ignore_nonleaf && n_nonleaf > 0 && n_leaf > 0);
-	rc = (n_nonleaf ? -2 : 0);
+	rc = process_deletions(p, handler, userdata, offsets, clean);
 
 cleanup:
-	if (entry) {
-		if (*entry_dn(entry))
-			fprintf(stderr, "Error at: %s\n", entry_dn(entry));
-		entry_free(entry);
-	}
-	if (cleanentry) entry_free(cleanentry);
 	if (key) free(key);
 
 	if (syntax_error_position)
