@@ -1,4 +1,5 @@
-/* Copyright (c) 2003,2004,2005 David Lichteblau
+/* -*- show-trailing-whitespace: t; indent-tabs: t -*-
+ * Copyright (c) 2003,2004,2005,2006 David Lichteblau
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -16,37 +17,6 @@
  */
 #include "common.h"
 
-static void
-print_entry_message(FILE *s, LDAP *ld, LDAPMessage *entry, int key)
-{
-	char *dn, *ad;
-	BerElement *ber;
-
-	fprintf(s, "\n%d ", key);
-	fputs(dn = ldap_get_dn(ld, entry), s);
-	ldap_memfree(dn);
-	fputc('\n', s);
-
-	for (ad = ldap_first_attribute(ld, entry, &ber);
-	     ad;
-	     ad = ldap_next_attribute(ld, entry, ber))
-	{
-		struct berval **values = ldap_get_values_len(ld, entry, ad);
-		struct berval **ptr;
-
-		if (!values) continue;
-		for (ptr = values; *ptr; ptr++) {
-			fputs(ad, s);
-			print_attrval(s, (*ptr)->bv_val, (*ptr)->bv_len);
-			fputc('\n', s);
-		}
-		ldap_memfree(ad);
-		ldap_value_free_len(values);
-	}
-	ber_free(ber, 0);
-	if (ferror(s)) syserr();
-}
-
 static int
 get_ws_col(void)
 {
@@ -74,7 +44,7 @@ update_progress(LDAP *ld, int n, LDAPMessage *entry)
 			return;
 		usec = tv.tv_usec;
 	}
-	
+
 	putchar('\r');
 	for (i = 0; i < cols; i++) putchar(' ');
 
@@ -96,7 +66,7 @@ handle_result(LDAP *ld, LDAPMessage *result, int start, int n,
         int err;
         char *matcheddn;
         char *text;
-	
+
         rc = ldap_parse_result(ld, result, &err, &matcheddn, &text, 0, 0, 0);
         if (rc) ldaperr(ld, "ldap_parse_result");
 
@@ -143,15 +113,54 @@ log_reference(LDAP *ld, LDAPMessage *reference, FILE *s)
 	ldap_value_free(refs);
 }
 
+static tentroid *
+entroid_set_message(LDAP *ld, tentroid *entroid, LDAPMessage *entry)
+{
+	struct berval **values = ldap_get_values_len(ld, entry, "objectClass");
+	struct berval **ptr;
+
+	if (!values || !*values)
+		return 0;
+
+	entroid_reset(entroid);
+	for (ptr = values; *ptr; ptr++) {
+		struct berval *value = *ptr;
+		LDAPObjectClass *cls
+			= entroid_request_class(entroid, value->bv_val);
+		if (!cls) {
+			g_string_append(entroid->comment, "# ERROR: ");
+			g_string_append(entroid->comment, entroid->error->str);
+			ldap_value_free_len(values);
+			return entroid;
+		}
+	}
+	ldap_value_free_len(values);
+
+	if (compute_entroid(entroid) == -1) {
+		g_string_append(entroid->comment, "# ERROR: ");
+		g_string_append(entroid->comment, entroid->error->str);
+		return entroid;
+	}
+	return entroid;
+}
+
 static void
 search_subtree(FILE *s, LDAP *ld, GArray *offsets, char *base,
-	       cmdline *cmdline, LDAPControl **ctrls, int notty)
+	       cmdline *cmdline, LDAPControl **ctrls, int notty, int ldif,
+	       tschema *schema)
 {
 	int msgid;
 	LDAPMessage *result, *entry;
 	int start = offsets->len;
 	int n = start;
 	long offset;
+	tentroid *entroid;
+	tentroid *e;
+
+	if (schema)
+		entroid = entroid_new(schema);
+	else
+		entroid = 0;
 
 	if (ldap_search_ext(
 		    ld, base,
@@ -169,9 +178,17 @@ search_subtree(FILE *s, LDAP *ld, GArray *offsets, char *base,
 			offset = ftell(s);
 			if (offset == -1 && !notty) syserr();
 			g_array_append_val(offsets, offset);
-			print_entry_message(s, ld, entry, n);
+			if (entroid)
+				e = entroid_set_message(ld, entroid, entry);
+			else
+				e = 0;
+			if (ldif)
+				print_ldif_message(
+					s, ld, entry, notty ? -1 : n, e);
+			else
+				print_ldapvi_message(s, ld, entry, n, e);
 			n++;
-			if (cmdline->progress && !notty)
+			if (!cmdline->quiet && !notty)
 				update_progress(ld, n, entry);
 			ldap_msgfree(entry);
 			break;
@@ -184,7 +201,7 @@ search_subtree(FILE *s, LDAP *ld, GArray *offsets, char *base,
 				update_progress(ld, n, 0);
 				putchar('\n');
 			}
-			handle_result(ld, result, start, n, cmdline->progress,
+			handle_result(ld, result, start, n, !cmdline->quiet,
 				      notty);
 			n = -1;
 			ldap_msgfree(result);
@@ -192,37 +209,57 @@ search_subtree(FILE *s, LDAP *ld, GArray *offsets, char *base,
 		default:
 			abort();
 		}
+	if (entroid)
+		entroid_free(entroid);
 }
 
 GArray *
-search(FILE *s, LDAP *ld, cmdline *cmdline, LDAPControl **ctrls, int notty)
+search(FILE *s, LDAP *ld, cmdline *cmdline, LDAPControl **ctrls, int notty,
+       int ldif)
 {
 	GArray *offsets = g_array_new(0, 0, sizeof(long));
 	GPtrArray *basedns = cmdline->basedns;
 	int i;
+	tschema *schema;
+
+	if (cmdline->schema_comments) {
+		schema = schema_new(ld);
+		if (!schema) {
+			fputs("Error: Failed to read schema, giving up.",
+			      stderr);
+			exit(1);
+		}
+	} else
+		schema = 0;
 
 	if (basedns->len == 0)
-		search_subtree(s, ld, offsets, 0, cmdline, ctrls, notty);
+		search_subtree(s, ld, offsets, 0, cmdline, ctrls, notty, ldif,
+			       schema);
 	else
 		for (i = 0; i < basedns->len; i++) {
 			char *base = g_ptr_array_index(basedns, i);
-			if (cmdline->progress && (basedns->len > 1))
+			if (!cmdline->quiet && (basedns->len > 1))
 				fprintf(stderr, "Searching in: %s\n", base);
-			search_subtree(
-				s, ld, offsets, base, cmdline, ctrls, notty);
+			search_subtree(s, ld, offsets, base, cmdline, ctrls,
+				       notty, ldif, schema);
 		}
 
 	if (!offsets->len) {
-		if (!cmdline->progress) /* if not printed already... */
-			fputs("No search results.  ", stderr);
-		fputs("(Maybe use --add instead?)\n", stderr);
+		if (!cmdline->noninteractive) {
+			if (cmdline->quiet) /* if not printed already... */
+				fputs("No search results.  ", stderr);
+			fputs("(Maybe use --add or --discover instead?)\n",
+			      stderr);
+		}
 		exit(0);
 	}
 
+	if (schema)
+		schema_free(schema);
 	return offsets;
 }
 
-static LDAPMessage *
+LDAPMessage *
 get_entry(LDAP *ld, char *dn, LDAPMessage **result)
 {
 	LDAPMessage *entry;
@@ -251,51 +288,3 @@ discover_naming_contexts(LDAP *ld, GPtrArray *basedns)
 	}
 	ldap_msgfree(result);
 }
-
-void
-get_schema(LDAP *ld, GPtrArray *objectclasses, GPtrArray *attributetypes)
-{
-	LDAPMessage *result, *entry;
-	char **values;
-	char *subschema_dn;
-	int code;
-	const char *errp;
-
-	entry = get_entry(ld, "", &result);
-	values = ldap_get_values(ld, entry, "subschemaSubentry");
-	if (!values) {
-		ldap_msgfree(result);
-		return;
-	}
-	subschema_dn = xdup(*values);
-	ldap_value_free(values);
-	ldap_msgfree(result);
-
-	entry = get_entry(ld, subschema_dn, &result);
-	free(subschema_dn);
-	values = ldap_get_values(ld, entry, "objectClasses");
-	if (values) {
-		char **ptr = values;
-		for (ptr = values; *ptr; ptr++) {
-			struct ldap_objectclass *cls
-				= ldap_str2objectclass(
-					*ptr, &code, &errp, 0);
-			if (!cls) yourfault(ldap_scherr2str(code));
-			g_ptr_array_add(objectclasses, cls);
-		}
-		ldap_value_free(values);
-	}
-	values = ldap_get_values(ld, entry, "attributeTypes");
-	if (values) {
-		char **ptr = values;
-		for (ptr = values; *ptr; ptr++) {
-			struct ldap_attributetype *at
-				= ldap_str2attributetype(
-					*ptr, &code, &errp, 0);
-			if (!at) yourfault(ldap_scherr2str(code));
-			g_ptr_array_add(attributetypes, at);
-		}
-		ldap_value_free(values);
-	}
-	ldap_msgfree(result);
-}
